import tomli
import os
import platformdirs
import argparse
import getpass
import sys

default_folder = platformdirs.user_config_dir("letterboxd_stats", getpass.getuser())

CONFIG_DEFAULTS = {
    "CLI": {
        "poster_columns": 180,
        "ascending": False,
    },
    "TMDB": {
        "get_list_runtimes": False,
    },
}

def load_config(file_path, defaults):
    # Load the TOML file if it exists
    if os.path.exists(file_path):
        with open(file_path, "rb") as f:
            user_config = tomli.load(f)
    else:
        user_config = {}

    # Merge user_config with defaults (user_config takes precedence)
    return merge_dicts(defaults, user_config)

def merge_dicts(defaults, overrides):
    """Recursively merge two dictionaries."""
    merged = defaults.copy()
    for key, value in overrides.items():
        if isinstance(value, dict) and key in merged and isinstance(merged[key], dict):
            merged[key] = merge_dicts(merged[key], value)
        else:
            merged[key] = value
    return merged

parser = argparse.ArgumentParser(
    prog="Letterboxd Stats",
    description="CLI tool to display Letterboxd statistics",
)
parser.add_argument("-s", "--search", help="Search for a director")
parser.add_argument("-S", "--search-film", help="Search for a film.")
parser.add_argument(
    "-d",
    "--download",
    help="Download letterboxd data from your account",
    action="store_true",
)
parser.add_argument("-W", "--watchlist", help="show watchlist", action="store_true")
parser.add_argument("-D", "--diary", help="show diary", action="store_true")
parser.add_argument("-R", "--ratings", help="show ratings", action="store_true")
parser.add_argument("-L", "--lists", help="show lists", action="store_true")
parser.add_argument("-l", "--limit", help="limit the number of items of your wishlist/diary", type=int)
parser.add_argument("-c", "--config_folder", help="Specifiy the folder of your config.toml file")

if len(sys.argv)==1:
    parser.print_help(sys.stderr)
    sys.exit(1)

args = parser.parse_args()

folder = args.config_folder or default_folder
config_path = os.path.abspath(os.path.join(folder, "config.toml"))
if not os.path.exists(config_path):
    raise FileNotFoundError(
        f"Found no configuration file in {config_path}. "
        + "Please, add a config.toml in that folder or specify a custom one with the -c command."
    )
    
config = load_config(config_path, CONFIG_DEFAULTS)




import pandas as pd
from pandarallel import pandarallel
import numpy as np
from letterboxd_stats import cli
from letterboxd_stats.web_scraper import get_tmdb_id
from letterboxd_stats import tmdb
import os
from letterboxd_stats import config
from tqdm import tqdm

tqdm.pandas(desc="Fetching ids...")
pandarallel.initialize(progress_bar=False, verbose=1)


def check_if_watched(df: pd.DataFrame, row: pd.Series) -> bool:
    """watched.csv hasn't the TMDB id, so comparison can be done only by title.
    This creates the risk of mismatch when two movies have the same title. To avoid this,
    we must retrieve the TMDB id of the watched movie.
    """

    if row["Title"] in df["Name"].values:
        watched_films_same_name = df[df["Name"] == row["Title"]]
        for _, film in watched_films_same_name.iterrows():
            film_id = get_tmdb_id(film["Letterboxd URI"])
            if film_id == row.name:
                return True
    return False


def read_watched_films(df: pd.DataFrame, path: str, name: str) -> pd.DataFrame:
    """Check which film of a director you have seen. Add a column to show on the CLI."""

    df_profile = pd.read_csv(path)
    df.insert(
        0,
        "watched",
        np.where(
            [check_if_watched(df_profile, row) for _, row in df.iterrows()],
            "[X]",
            "[ ]",
        ),
    )
    df["Release Date"] = pd.to_datetime(df["Release Date"])
    df.sort_values(by="Release Date", inplace=True)
    cli.render_table(df, name)
    return df


def select_film_of_person(df: pd.DataFrame) -> pd.Series | None:
    movie_id = cli.select_movie(df["Title"], df.index.to_series())
    if movie_id is None:
        return None
    movie_row = df.loc[movie_id]
    return movie_row


def get_list_name(path: str) -> str:
    df = pd.read_csv(path, header=1)
    return df["Name"].iloc[0]


def open_list(path: str, limit: int, acending: bool) -> str:
    """Select a list from the saved ones."""

    list_names = {
        get_list_name(os.path.join(path, letterboxd_list)): letterboxd_list for letterboxd_list in os.listdir(path)
    }
    name = cli.select_list(sorted(list(list_names.keys())))
    return open_file("Lists", os.path.join(path, list_names[name]), limit, acending, header=3)


def open_file(filetype: str, path: str, limit, ascending, header=0) -> str:
    """There are some operations that are the same for all the .csv files. So isolate those similar operations,
    and then we proceed to perform the particular operation for a certain file (watchlist, list, diary...).
    FILE_OPERATIONS selects those particular operations according to the file we opened. Mainly they do
    ordering and column filtering operations.
    """

    df = pd.read_csv(path, header=header)
    df.rename(columns={"Name": "Title", "Letterboxd URI": "Url"}, inplace=True)
    df["Year"] = df["Year"].fillna(0).astype(int)
    df = FILE_OPERATIONS[filetype](df, ascending)
    if limit is not None:
        df = df.iloc[:limit, :]
    cli.render_table(df, filetype)
    return cli.select_movie(df["Title"], df["Url"])


def _show_lists(df: pd.DataFrame, ascending: bool) -> pd.DataFrame:
    ratings_path = os.path.expanduser(os.path.join(config["root_folder"], "static", "ratings.csv"))
    df_ratings = pd.read_csv(ratings_path)
    df_ratings.rename(columns={"Letterboxd URI": "URL"}, inplace=True)
    df = df.merge(df_ratings[["URL", "Rating"]], on="URL", how="inner")
    df.rename(columns={"URL": "Url"}, inplace=True)
    df = df.drop("Description", axis=1)
    df["Rating"] = df["Rating"].astype(float)
    sort_column = cli.select_value(df.columns.values.tolist(), "Select the order of your diary entries:")
    df.sort_values(by=sort_column, ascending=ascending, inplace=True)
    avg = {"Rating Mean": "{:.2f}".format(df["Rating"].mean())}
    if config["TMDB"]["get_list_runtimes"] is True:
        ids = df["Url"].parallel_map(get_tmdb_id)
        df["Duration"] = ids.parallel_map(lambda id: tmdb.get_film_duration(id))  # type: ignore
        avg["Time-weighted Rating Mean"] = "{:.2f}".format(
            ((df["Duration"] / df["Duration"].sum()) * df["Rating"]).sum()
        )
    cli.print_film(avg, expand=False)
    return df


def _show_watchlist(df: pd.DataFrame, ascending: bool) -> pd.DataFrame:
    sort_column = cli.select_value(
        df.columns.values.tolist() + ["Shuffle"],
        "Select the order of your watchlist entries:",
    )
    if sort_column == "Shuffle":
        df = df.sample(frac=1)
    else:
        df.sort_values(by=sort_column, ascending=ascending, inplace=True)
    return df


def _show_diary(df: pd.DataFrame, ascending: bool) -> pd.DataFrame:
    df["Watched Date"] = pd.to_datetime(df["Watched Date"])
    sort_column = cli.select_value(df.columns.values.tolist(), "Select the order of your diary entries:")
    df.sort_values(by=sort_column, ascending=ascending, inplace=True)
    df = df.drop(["Rewatch", "Tags"], axis=1)
    return df


def _show_ratings(df: pd.DataFrame, ascending: bool) -> pd.DataFrame:
    df["Date"] = pd.to_datetime(df["Date"])
    sort_column = cli.select_value(df.columns.values.tolist(), "Select the order of your ratings:")
    df.sort_values(by=sort_column, ascending=ascending, inplace=True)
    if sort_column == "Rating":
        options = df["Rating"].unique().tolist()
        rating_range = cli.select_range(options=options)
        df = df[df["Rating"].isin(rating_range)]
    return df


FILE_OPERATIONS = {
    "Diary": _show_diary,
    "Watchlist": _show_watchlist,
    "Ratings": _show_ratings,
    "Lists": _show_lists,
}



import os
from zipfile import ZipFile
from letterboxd_stats import config
from letterboxd_stats import cli
import requests
from lxml import html
import shelve

URL = "https://letterboxd.com"
LOGIN_PAGE = URL + "/user/login.do"
DATA_PAGE = URL + "/data/export"
ADD_DIARY_URL = URL + "/s/save-diary-entry"
MOVIE_OPERATIONS = {
    "Add to diary": "add_film_diary",
    "Add to watchlist": "add_watchlist",
    "Remove from watchlist": "remove_watchlist",
}
OPERATIONS_URLS = {
    "search": lambda s: f"/s/search/{s}/",
    "diary": lambda s: f"/csi/film/{s}/sidebar-user-actions/?esiAllowUser=true",
    "add_watchlist": lambda s: f"/film/{s}/add-to-watchlist/",
    "remove_watchlist": lambda s: f"/film/{s}/remove-from-watchlist/",
    "film_page": lambda s: f"/film/{s}",
}

cache_path = os.path.expanduser(os.path.join(config["root_folder"], "static", "cache.db"))


class Downloader:
    def __init__(self):
        self.session = requests.Session()
        # get home page to set cookies in the session.
        self.session.get(URL)

    def login(self):
        request_payload = {
            "username": config["Letterboxd"]["username"],
            "password": config["Letterboxd"]["password"],
            "__csrf": self.session.cookies.get("com.xk72.webparts.csrf"),
        }
        res = self.session.post(LOGIN_PAGE, data=request_payload)
        if res.json()["result"] != "success":
            raise ConnectionError("Impossible to login")

    def download_stats(self):
        """Download and extract data of the import/export section.
        .CSV file will be extracted in the folder specified in the config file."""

        res = self.session.get(DATA_PAGE)
        if res.status_code != 200 or "application/zip" not in res.headers["Content-Type"]:
            raise ConnectionError(f"Impossible to download data. Response headers:\n{res.headers}")
        print("Data download successful.")
        filename = res.headers["content-disposition"].split()[-1].split("=")[-1]
        path = os.path.expanduser(os.path.join(config["root_folder"], "static"))
        if not os.path.exists(path):
            os.makedirs(path)
        archive = os.path.join(path, filename)
        with open(archive, "wb") as f:
            f.write(res.content)
        with ZipFile(archive, "r") as zip:
            zip.extractall(path)
        os.remove(archive)

    def add_film_diary(self, title: str):
        payload = cli.add_film_questions()
        url = create_movie_url(title, "diary")
        res = self.session.get(url)
        if res.status_code != 200:
            raise ConnectionError("It's been impossible to retireve the Letterboxd page")
        movie_page = html.fromstring(res.text)
        # Not the TMDB id, but the Letterboxd ID to use to add the movie to diary.
        # Reference: https://letterboxd.com/film/seven-samurai/
        letterboxd_film_id = movie_page.get_element_by_id("frm-sidebar-rating").get("data-rateable-uid").split(":", 1)[1]
        payload["filmId"] = letterboxd_film_id
        payload["__csrf"] = self.session.cookies.get("com.xk72.webparts.csrf")
        res = self.session.post(ADD_DIARY_URL, data=payload)
        if not (res.status_code == 200 and res.json()["result"] is True):
            raise ConnectionError(f"Add to diary request failed.")
        print("The movie was added to your diary.")

    def add_watchlist(self, title: str):
        url = create_movie_url(title, "add_watchlist")
        res = self.session.post(url, data={"__csrf": self.session.cookies.get("com.xk72.webparts.csrf")})
        if not (res.status_code == 200 and res.json()["result"] is True):
            raise ConnectionError("Add to watchlist request failed.")
        print("Added to your watchlist.")

    def remove_watchlist(self, title: str):
        url = create_movie_url(title, "remove_watchlist")
        res = self.session.post(url, data={"__csrf": self.session.cookies.get("com.xk72.webparts.csrf")})
        if not (res.status_code == 200 and res.json()["result"] is True):
            raise ConnectionError("Remove from watchlist request failed.")
        print("Removed to your watchlist.")

    def perform_operation(self, answer: str, link: str):
        """Depending on what the user has chosen, add to diary, add/remove watchlist."""

        getattr(self, MOVIE_OPERATIONS[answer])(link)


def create_movie_url(title: str, operation: str) -> str:
    return URL + OPERATIONS_URLS[operation](title)


def _get_tmdb_id_from_web(link: str, is_diary: bool) -> int:
    """Scraping the TMDB link from a Letterboxd film page.
    Inpect this HTML for reference: https://letterboxd.com/film/seven-samurai/
    """

    res = requests.get(link)
    movie_page = html.fromstring(res.text)
    # Diary links sends you to a different page with no link to TMDB. Redirect to the actual page.
    if is_diary:
        title_link = movie_page.xpath("//span[@class='film-title-wrapper']/a")
        if len(title_link) == 0:
            raise ValueError("No movie link found.")
        movie_link = title_link[0]
        movie_url = URL + movie_link.get("href")
        movie_page = html.fromstring(requests.get(movie_url).text)
    tmdb_link = movie_page.xpath("//a[@data-track-action='TMDb']")
    if len(tmdb_link) == 0:
        raise ValueError("No Movie link found")
    id = tmdb_link[0].get("href").split("/")[-2]
    return int(id)


def get_tmdb_id(link: str, is_diary=False) -> int | None:
    """Find the TMDB id from a letterboxd page.

    A link to a Letterboxd film usually starts with either https://letterboxd.com/
    or https://boxd.it/ (usually all .csv files have this prefix). We structure the cache dict accordingly.
    The cache is meant to avoid bottleneck of constantly retrieving the Id from an HTML page.
    """

    tmdb_id_cache = shelve.open(cache_path, writeback=False, protocol=5)
    prefix, key = link.rsplit("/", 1)
    if prefix in tmdb_id_cache and key in tmdb_id_cache[prefix]:
        id = tmdb_id_cache[prefix][key]
    else:
        try:
            id = _get_tmdb_id_from_web(link, is_diary)
            prefix_dict = tmdb_id_cache.get(prefix) or {}
            prefix_dict[key] = id
            tmdb_id_cache[prefix] = prefix_dict
        except ValueError as e:
            print(e)
            id = None
    tmdb_id_cache.close()
    return id


def select_optional_operation() -> str:
    return cli.select_value(["Exit"] + list(MOVIE_OPERATIONS.keys()), "Select operation:")


def get_lb_title(title: str, allow_selection=False) -> str:
    """Search a movie a get its Letterboxd link.
    For reference: https://letterboxd.com/search/seven+samurai/?adult
    """

    search_url = create_movie_url(title, "search")
    res = requests.get(search_url)
    if res.status_code != 200:
        raise ConnectionError("It's been impossible to retireve the Letterboxd page")
    search_page = html.fromstring(res.text)
    # If we want to select movies from the seach page, get more data to print the selection prompt.
    if allow_selection:
        movie_list = search_page.xpath("//div[@class='film-detail-content']")
        if len(movie_list) == 0:
            raise ValueError(f"No film found with search query {title}")
        title_years_directors_links = {}
        for movie in movie_list:
            title = movie.xpath("./h2/span/a")[0].text.rstrip()
            director = director[0].text if len(director := movie.xpath("./p/a")) > 0 else ""
            year = f"({year[0].text}) " if len(year := movie.xpath("./h2/span//small/a")) > 0 else ""
            link = movie.xpath("./h2/span/a")[0].get("href")
            title_years_directors_links[f"{title} {year}- {director}"] = link
        selected_film = cli.select_value(list(title_years_directors_links.keys()), "Select your film")
        title_url = title_years_directors_links[selected_film].split("/")[-2]
    else:
        title_url = search_page.xpath("//span[@class='film-title-wrapper']/a")[0].get("href").split("/")[-2]
    return title_url



from letterboxd_stats import tmdb
from letterboxd_stats import data
from letterboxd_stats import web_scraper as ws
import os
from letterboxd_stats import args, config

DATA_FILES = {"Watchlist": "watchlist.csv", "Diary": "diary.csv", "Ratings": "ratings.csv", "Lists": "lists"}


def try_command(command, args):
    try:
        command(*args)
    except Exception as e:
        print(e)


def check_path(path: str):
    if not os.path.exists(path):
        raise FileNotFoundError(
            f"No Letterboxd data was found in {path}. Make sure the path is correct or run -d to download your data"
        )


def download_data():
    """Download exported data you find in the import/esport section of your Letterboxd profile"""

    downloader = ws.Downloader()
    downloader.login()
    downloader.download_stats()


def search_person(args_search: str):
    """Search for a director, list his/her movies and check if you have watched them."""

    df, name = tmdb.get_person(args_search)
    path = os.path.expanduser(os.path.join(config["root_folder"], "static", "watched.csv"))
    check_path(path)
    df = data.read_watched_films(df, path, name)
    movie = data.select_film_of_person(df)
    # We want to print the link of the selected movie. This has to be retrived from the search page.
    while movie is not None:
        search_film_query = f"{movie['Title']} {movie['Release Date'].year}"  # type: ignore
        title_url = ws.get_lb_title(search_film_query)
        tmdb.get_movie_detail(int(movie.name), ws.create_movie_url(title_url, "film_page"))  # type: ignore
        movie = data.select_film_of_person(df)


def search_film(args_search_film: str):
    title_url = ws.get_lb_title(args_search_film, True)
    film_url = ws.create_movie_url(title_url, "film_page")
    tmdb.get_movie_detail(ws.get_tmdb_id(film_url), film_url)  # type: ignore
    answer = ws.select_optional_operation()
    if answer != "Exit":
        downloader = ws.Downloader()
        downloader.login()
        downloader.perform_operation(answer, title_url)


def display_data(args_limit: int, args_ascending: bool, data_type: str):
    """Load and show on the CLI different .csv files that you have downloaded with the -d flag."""

    path = os.path.expanduser(os.path.join(config["root_folder"], "static", DATA_FILES[data_type]))
    check_path(path)
    letterboxd_url = (
        data.open_file(data_type, path, args_limit, args_ascending)
        if data_type != "Lists"
        else data.open_list(path, args_limit, args_ascending)
    )
    # If you select a movie, show its details.
    if letterboxd_url is not None:
        id = ws.get_tmdb_id(letterboxd_url, data_type == "diary")
        if id is not None:
            tmdb.get_movie_detail(id, letterboxd_url)


def main():
    if args.download:
        try_command(download_data, ())
    if args.search:
        try_command(search_person, (args.search,))
    if args.search_film:
        try_command(search_film, (args.search_film,))
    if args.watchlist:
        try_command(display_data, (args.limit, config["CLI"]["ascending"], "Watchlist"))
    if args.diary:
        try_command(display_data, (args.limit, config["CLI"]["ascending"], "Diary"))
    if args.ratings:
        try_command(display_data, (args.limit, config["CLI"]["ascending"], "Ratings"))
    if args.lists:
        try_command(display_data, (args.limit, config["CLI"]["ascending"], "Lists"))


if __name__ == "__main__":
    main()



from typing import Any, Tuple
from tmdbv3api import TMDb, Person, Movie, Search
from tmdbv3api.exceptions import TMDbException
import pandas as pd
from pandarallel import pandarallel
from tmdbv3api.objs.account import AsObj
from letterboxd_stats import cli
from letterboxd_stats import config

tmdb = TMDb()
tmdb.api_key = config["TMDB"]["api_key"]
person = Person()
movie = Movie()
search = Search()
pandarallel.initialize(progress_bar=False, verbose=1)


def get_person(name: str) -> Tuple[pd.DataFrame, str]:
    """Search the director with the TMDB api. Get all the movies.
    https://developer.themoviedb.org/reference/person-details
    https://developer.themoviedb.org/reference/person-movie-credits
    """

    print(f"Searching for '{name}'")
    search_results = search.people({"query": name})
    names = [result.name for result in search_results]  # type: ignore
    if len(names) == 0:
        raise Exception("No results for your search")
    result_index = cli.select_search_result(names)  # type: ignore
    search_result = search_results[result_index]
    p = person.details(search_result["id"])
    known_for_department = p["known_for_department"]
    movie_credits = person.movie_credits(search_result["id"])
    list_of_films = [
        {
            "Id": m.id,
            "Title": m.title,
            "Release Date": m.release_date,
            "Department": m.department,
        }
        for m in movie_credits["crew"]
    ]
    if len(list_of_films) == 0:
        raise ValueError("The selected person doesn't have any film.")
    df = pd.DataFrame(list_of_films).set_index("Id")
    department = cli.select_value(
        df["Department"].unique(), f"Select a department for {p['name']}", known_for_department
    )
    df = df[df["Department"] == department]
    df = df.drop("Department", axis=1)
    # person.details provides movies without time duration. If the user wants<S-D-A>
    # (since this slows down the process) get with the movie.details API.
    if config["TMDB"]["get_list_runtimes"] is True:
        df["Duration"] = df.index.to_series().parallel_map(get_film_duration)  # type: ignore
    return df, p["name"]


def get_movie(movie_query: str) -> Any | AsObj:
    print(f"Searching for movie '{movie_query}'")
    search_results = search.movies({"query": movie_query})
    titles = [f"{result.title} ({result.release_date})" for result in search_results]  # type: ignore
    if len(titles) == 0:
        raise Exception("No results for your search")
    result_index = cli.select_search_result(titles)  # type: ignore
    movie_id = search_results[result_index]["id"]
    get_movie_detail(movie_id)
    return search_results[result_index]


def get_movie_detail(movie_id: int, letterboxd_url=None):
    movie_details = movie.details(movie_id)
    poster = movie_details.get("poster_path")
    if poster is not None:
        cli.download_poster(poster)
    selected_details = {
        "Title": movie_details["title"],
        "Original Title": movie_details["original_title"],
        "Runtime": movie_details["runtime"],
        "Overview": movie_details["overview"],
        "Release Date": movie_details["release_date"],
    }
    if letterboxd_url:
        selected_details["Letterboxd Url"] = letterboxd_url
    cli.print_film(selected_details)


def get_film_duration(tmdb_id: int) -> int:
    """Get film duration from the TMDB api.
    https://developer.themoviedb.org/reference/movie-details
    """

    try:
        runtime = movie.details(tmdb_id).runtime  # type: ignore
    except TMDbException:
        runtime = 0
    return runtime
